import { Nag, VCardNagAttributes } from './errors.js';
import { NonEmptyArray } from './nonEmptyArray.js';
import { VCard4, VCardParameters } from './vcard4Types.js';
export declare type LineAttributes = {
    property: string;
    field: string;
    entireLine: string;
    shortenedLine: string;
};
export declare type ParsedVCards = {
    vCards?: NonEmptyArray<VCard4>;
    nags?: NonEmptyArray<Nag<undefined>>;
};
export declare type PartialVCard = Partial<Omit<VCard4, 'nags' | 'unparseable'>> & {
    nags: Nag<VCardNagAttributes>[];
    unparseable: string[];
    didNotStartWithBEGIN?: boolean;
};
export declare type TransitionVCard = Partial<Omit<VCard4, 'nags' | 'unparseable'>> & {
    nags?: Nag<VCardNagAttributes>[];
    unparseable?: string[];
    didNotStartWithBEGIN?: boolean;
};
/**
 * Parse an RFC 6350 (multi-)vCard input into an array and errors.
 * @param vcf Multiline string of potentially multiple vCards
 * @param keepErrors Whether very bad vCards should be deleted
 * @returns Array of vCards with metadata; array of global errors
 */
export declare function parseVCards(vcf: string, keepDefective?: boolean): ParsedVCards;
/**
 * Parse a line and extend the partial vCard accordingly.
 * @param vCardInProgress The vCard to add the line to, if possible
 * @param line The new line
 */
export declare function parseLine(vCardInProgress: PartialVCard, line: string): void;
/**
 * Scan a group or property name.
 * @param line The entire line
 * @param start The offset at which to start scanning
 * @returns The scanned name and end position in upper case, or throws an error
 */
export declare function scanPropertyOrGroup(line: string, start: number, nags: Nag<VCardNagAttributes>[]): {
    name: string;
    end: number;
} | null;
declare type GroupedProperty = {
    group?: Uppercase<string>;
    property: Uppercase<string>;
    end: number;
};
/**
 * Extract (group and) property names.
 * @param line The line to be scanned (from start)
 * @returns (Optional) group, property (both in uppercase), and scanning end. Or null on error.
 */
export declare function extractProperty(line: string, nags: Nag<VCardNagAttributes>[]): GroupedProperty | null;
/**
 * Scan a lone parameter value, without unescaping it
 * @param line The line to be parsed
 * @param start The offset to start parsing at (*after* the equals sign!)
 * @returns The parsed value and the end offset of it
 */
export declare function scanParamRawValue(line: string, start: number): {
    value: string;
    end: number;
};
/**
 * Scan the potentially multiple values of this parameter.
 * @param line The line currently being parsed
 * @param start The offset to start parsing at
 * @param property The name of the property this parameter belongs to
 * @param parameter The name of the parameter
 * @param nags The list of all complaints
 * @param singleValue Do not treat commas as separator
 * @returns A string[] value and a parsing end, or null
 */
export declare function scanParamValues(line: string, start: number, property: string, parameter: string, nags: Nag<VCardNagAttributes>[], singleValue?: boolean): {
    value: string[];
    end: number;
} | null;
/**
 * Scan a lone parameter value, unescaping it
 * @param line The line to be parsed
 * @param start The offset to start parsing at (the equals sign)
 * @param property The name of the property this parameter belongs to
 * @param parameter The name of the parameter
 * @param nags The list of all complaints
 * @returns The parsed value and the end offset of it
 */
export declare function scanParamValue(line: string, start: number, property: string, parameter: string, nags: Nag<VCardNagAttributes>[]): {
    value: string;
    end: number;
} | null;
/**
 * Parse the parameters. Used after scanning the property or group. What remains should be the value.
 * @param line The vCard line to be analyzed
 * @param start The offset where to start parsing (the equals sign)
 * @returns A record of VCardParameters, and the end offset.
 */
export declare function parseParameters(line: string, start: number, nags: Nag<VCardNagAttributes>[], property: string): {
    parameters: VCardParameters;
    end: number;
} | null;
export {};
