import { errorKeys } from './errorCodes.js';
import { NonEmptyArray } from './nonEmptyArray.js';
export declare type TypeOf<T extends RuntimeTypeAnnotation<any>> = T['type'];
export declare type Flatten<T extends object> = {
    [k in keyof T]: T[k];
};
/**
 * Creates a subtype of the base schema, where only the keys are included
 * whose type is compatible with Condition.
 *
 * E.g.
 *   SubType<{BEGIN: string; END: string; NICKNAME: string[];}, string>
 * will result in
 *   {BEGIN: string; END: string;}
 *
 * Obtained from https://www.piotrl.net/typescript-condition-subset-types/
 */
export declare type SubType<Base, Condition> = Pick<Base, {
    [Key in keyof Base]: Base[Key] extends Condition ? Key : never;
}[keyof Base]>;
/**
 * Creates a subtype of the Base schema, where only the keys are included
 * whose RuntimeTypeAnnotation type is compatible with Condition.
 *
 * E.g.
 *   RTSubType<
 *     {BEGIN: StringType; END: StringType; NICKNAME: NonEmptyStringType;},
 *     string>
 * will result in
 *   {BEGIN: StringType; END: StringType;}
 */
export declare type RTSubType<Base extends Record<string, RuntimeTypeAnnotation<any>>, Condition> = Pick<Base, {
    [Key in keyof Base]: Base[Key]['type'] extends Condition ? Key : never;
}[keyof Base]>;
export declare type Name = 'string' | 'string[]' | 'string[+]' | 'number' | 'ADR' | 'N' | 'GENDER' | 'CLIENTPIDMAP';
export declare type ParseFunction<T> = (rawValue: string, errorCallback?: (error: errorKeys) => void) => T;
export declare class RuntimeTypeAnnotation<Type> {
    readonly type: Type;
    readonly name: Name;
    readonly parse: ParseFunction<Type>;
    constructor(name: Name, parse: ParseFunction<Type>);
}
export declare const StringType: RuntimeTypeAnnotation<string>;
export declare const NonEmptyStringArrayType: RuntimeTypeAnnotation<NonEmptyArray<string>>;
export declare const NumberType: RuntimeTypeAnnotation<number>;
export declare const ADRType: RuntimeTypeAnnotation<{
    postOfficeBox?: NonEmptyArray<string> | undefined;
    extendedAddress?: NonEmptyArray<string> | undefined;
    streetAddress?: NonEmptyArray<string> | undefined;
    locality?: NonEmptyArray<string> | undefined;
    region?: NonEmptyArray<string> | undefined;
    postalCode?: NonEmptyArray<string> | undefined;
    countryName?: NonEmptyArray<string> | undefined;
}>;
export declare const NType: RuntimeTypeAnnotation<{
    familyNames: NonEmptyArray<string>;
    givenNames: NonEmptyArray<string>;
    additionalNames: NonEmptyArray<string>;
    honorificPrefixes: NonEmptyArray<string>;
    honorificSuffixes: NonEmptyArray<string>;
}>;
export declare const GENDERType: RuntimeTypeAnnotation<{
    sex: string;
    text?: string | undefined;
}>;
export declare const CLIENTPIDMAPType: RuntimeTypeAnnotation<{
    pidRef: number;
    uri: string;
}>;
/**
 * Creates a type guard function, which will check whether its parameter, the key,
 * exists in the schema. The schema must be an actual object (typically a const),
 * it's type is inferred by the generic.
 *
 * Asserts to the type system that key is in the schema. This cannot be done directly
 * (as type information is only available at compile time), therefore, we need to
 * infer the type information from the schema (this direction is possible) and
 * verify at runtime against the schema.
 * @param schema The schema to check against (an actual variable/const, not a type!)
 * @returns The type guard function.
 */
export declare function keyTypeGuard<T extends Record<string, any>>(schema: T): (key: string) => key is keyof T & string;
/**
 * WIP
 *
 * Creates a type guard function, which will check whether its parameter, the key,
 * exists in the RuntimeTypeAnnotation schema with the given type name.
 * The schema must be an actual object (typically a const),
 * and the generic T must be `keyof typeof schema`.
 *
 * Asserts to the type system that key is in T. This cannot be done directly
 * (as type information is only available at compile time), therefore, we need to
 * infer the type information from the schema (this direction is possible) and
 * verify at runtime against the schema.
 * @param schema The schema to check against (an actual variable/const, not a type!)
 * @returns The type guard function.
 */
